# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
MSTICPy authentication and secretesmodules.

This is sub-package authentication and configuration
modules and classes.

Some of the modules and classes here:

- azure_auth - az_connect function
- secret_settings - handles Azure Key Vault and keyring-stored secrets

"""
